# Copyright 2016 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import numpy as np
from .utils import RandomParameterSearch, PARAMS
from SparseArray import SparseArray
from .model import Ensemble, EvoDAG as model_EvoDAG, EvoDAGE
import collections
import os
from multiprocessing import Pool
import EvoDAG as evodag
from EvoDAG import EvoDAG
from .utils import tonparray
from .utils import Inputs
import time
import gzip
import json
import logging
import gc
import shutil
import pickle
DEFAULT_PARAMETERS = os.path.join(os.path.dirname(__file__),
                                  'conf', 'default_parameters.json')
try:
    from tqdm import tqdm
except ImportError:
    def tqdm(x, **kwargs):
        return x


def init_evodag(seed_args_X_y_test):
    seed, args, X, y, test, dirname = seed_args_X_y_test
    if dirname is not None:
        output = os.path.join(dirname, '%s.evodag' % seed)
        if os.path.isfile(output):
            with gzip.open(output) as fpt:
                try:
                    return pickle.load(fpt)
                except Exception:
                    pass
    m = EvoDAG(seed=seed, **args).fit(X, y, test_set=test)
    m = m.model()
    gc.collect()
    if dirname is not None:
        with gzip.open(output, 'w') as fpt:
            pickle.dump(m, fpt)
    return m


def rs_evodag(args_X_y):
    args, X, y = args_X_y
    rs = RandomParameterSearch
    fit = []
    init = time.time()
    for seed in range(3):
        try:
            evo = EvoDAG(seed=seed,
                         **rs.process_params(args)).fit(X, y)
            fit.append(evo.model().fitness_vs)
        except RuntimeError:
            fit.append(-np.inf)
    args['_time'] = time.time() - init
    gc.collect()
    return fit, args


def get_model_fitness(fname):
    with gzip.open(fname) as fpt:
        _ = pickle.load(fpt)
        return (fname, _.fitness_vs)


class CommandLine(object):
    def version(self):
        pa = self.parser.add_argument
        pa('--version',
           action='version', version='EvoDAG %s' % evodag.__version__)
        pa('--verbose', dest='verbose', default=logging.NOTSET, type=int)

    def output_file(self):
        self.parser.add_argument('-o', '--output-file',
                                 help='File to store the test set',
                                 dest='output_file',
                                 default=None,
                                 type=str)

    def ensemble(self):
        self.parser.add_argument('-n', '--ensemble-size',
                                 help='Ensemble size',
                                 dest='ensemble_size',
                                 default=30,
                                 type=int)

    def cores(self):
        self.parser.add_argument('-u', '--cpu-cores',
                                 help='Number of cores',
                                 dest='cpu_cores',
                                 default=1,
                                 type=int)

    def test_set(self):
        cdn = 'File containing the test set on csv.'
        self.parser.add_argument('-t', '--test_set',
                                 default=None, type=str,
                                 help=cdn)

    def init_params(self):
        pa = self.parser.add_argument
        g = self.parser.add_mutually_exclusive_group(required=True)
        g.add_argument('-C', '--classifier', dest='classifier',
                       help='The task is classification (default)',
                       default=True,
                       action="store_true")
        g.add_argument('-R', '--regressor', dest='regressor',
                       help='The task is regression',
                       action="store_true")
        pa('-e', '--early_stopping_rounds', dest='early_stopping_rounds',
           type=int,
           help='Early stopping rounds')
        pa('-p', '--popsize', dest='popsize',
           type=int, help='Population size')
        pa('-s', '--seed', dest='seed',
           default=0,
           type=int, help='Seed')
        pa('-j', '--json', dest='json',
           action="store_true",
           help='Whether the inputs are in json format',
           default=False)
        pa('--time-limit', dest='time_limit',
           help='Time limit in seconds', type=int)

    def training_set(self):
        cdn = 'File containing the training set on csv.'
        self.parser.add_argument('training_set',
                                 nargs='?',
                                 default=None,
                                 help=cdn)

    def parse_args(self):
        self.data = self.parser.parse_args()
        if hasattr(self.data, 'regressor') and self.data.regressor:
            self.data.classifier = False
        if hasattr(self.data, 'verbose'):
            logging.basicConfig()
            logger = logging.getLogger('EvoDAG')
            logger.setLevel(self.data.verbose)
            logger.info('Logging to: %s', self.data.verbose)
        self.main()

    def read_training_set(self):
        if self.data.training_set is None:
            return
        if not self.data.json:
            self.X, self.y = self.inputs.read_csv(self.data.training_set, self.data.output_dim)
            return True
        else:
            self.X, self.y = self.inputs.read_data_json(self.data.training_set)
            return True

    def read_test_set(self):
        if self.data.test_set is None:
            return False
        if not self.data.json:
            self.Xtest, _ = self.inputs.read_csv(self.data.test_set, 0)
            return True
        else:
            self.Xtest, _ = self.inputs.read_data_json(self.data.test_set)
            return True

    def get_model_file(self):
        if self.data.model_file is None:
            a = self.data.training_set.split('.')[0]
            self.data.model_file = a + '.evodag.gz'
        return self.data.model_file

    def get_output_file(self):
        if self.data.output_file is None:
            self.data.output_file = self.data.test_set + '.evodag.csv'
            # if self.data.json:
            #     self.data.output_file += '.json'
            # else:
            #     self.data.output_file += '.csv'
        return self.data.output_file

    def id2label(self, x):
        if not self.data.classifier:
            return x
        if len(self.label2id) == 0:
            return x
        i2w = dict([(i[1], i[0]) for i in self.label2id.items()])
        return [i2w[int(i)] for i in x]

    def main(self):
        pass


class CommandLineParams(CommandLine):
    def __init__(self):
        self.Xtest = None
        self.inputs = Inputs()
        self.word2id = self.inputs.word2id
        self.label2id = self.inputs.label2id
        self.parser = argparse.ArgumentParser(description="EvoDAG")
        self.training_set()
        self.init_params()
        self.optimize_parameters()
        self.cores()
        self.version()

    def optimize_parameters(self):
        cdn = '''Optimize parameters sampling
        N (734 by default) points from the parameter space'''
        self.parser.add_argument('-r', '--optimize-parameters',
                                 dest='optimize_parameters',
                                 default=734,
                                 type=int, help=cdn)
        cdn = 'File to store the fitness of the parameters explored'
        self.parser.add_argument('-P', '--parameters',
                                 dest='parameters',
                                 type=str,
                                 help=cdn)
        cdn = 'File containing the parameters values (json) to be explored'
        self.parser.add_argument('--parameters-values',
                                 dest='parameters_values',
                                 type=str,
                                 help=cdn)
        self.parser.add_argument('--output-dim',
                                 dest='output_dim',
                                 default=1,
                                 type=int,
                                 help="Output Dimension (default 1) use with multiple-outputs flag")
        self.parser.add_argument('--only-paramsfiles',
                                 help='Save the params to disk creating a directory',
                                 dest='do_nothing',
                                 action="store_true",
                                 default=False)

    def fs_type_constraint(self, params):
        fs_class = {}
        for x in EvoDAG()._function_set:
            fs_class[x.__name__] = x
        p_delete = []
        for x in params.keys():
            if x in fs_class:
                try:
                    if self.data.classifier:
                        flag = fs_class[x].classification
                    else:
                        flag = fs_class[x].regression
                    if not flag:
                        p_delete.append(x)
                except AttributeError:
                    pass
        for x in p_delete:
            del params[x]

    def if_type_contraint(self, params):
        import importlib
        unique = {}
        if 'input_functions' not in params:
            return
        input_functions = params['input_functions']
        R = []
        for inner in input_functions:
            r = []
            for f in inner:
                _ = importlib.import_module('EvoDAG.node')
                j = getattr(_, f)
                if self.data.classifier:
                    flag = j.classification
                else:
                    flag = j.regression
                if flag:
                    r.append(f)
            if len(r):
                key = ';'.join(r)
                if key not in unique:
                    R.append(r)
                    unique[key] = 1
        if len(R) == 1:
            R = R[0]
        params['input_functions'] = R

    def evolve(self, kw):
        if self.data.parameters_values:
            with open(self.data.parameters_values, 'r') as fpt:
                params = json.loads(fpt.read())
        else:
            params = PARAMS.copy()
        if len(kw):
            for k, v in kw.items():
                if k in params and v is not None:
                    params[k] = [v]
        self.fs_type_constraint(params)
        self.if_type_contraint(params)
        parameters = self.data.parameters
        if parameters is None:
            parameters = self.data.training_set + '.EvoDAGparams'
        npoints = self.data.optimize_parameters
        if isinstance(self.X, list):
            training_size = self.X[0].size()
        else:
            training_size = self.X.shape[0]
        rs = RandomParameterSearch(params=params,
                                   seed=self.data.seed,
                                   training_size=training_size,
                                   npoints=npoints)
        if self.data.do_nothing:
            os.mkdir(parameters)
            for k, x in enumerate(rs):
                fname = os.path.join(parameters, '%s_params.json' % k)
                with open(fname, 'w') as fpt:
                    fpt.write(json.dumps(x, sort_keys=True, indent=2))
            return
        if self.data.cpu_cores == 1:
            res = [rs_evodag((args, self.X, self.y))
                   for args in tqdm(rs, total=rs._npoints)]
        else:
            p = Pool(self.data.cpu_cores, maxtasksperchild=1)
            args = [(args, self.X, self.y) for args in rs]
            res = [x for x in tqdm(p.imap_unordered(rs_evodag, args),
                                   total=len(args))]
            p.close()
        [x[1].update(dict(fitness=x[0])) for x in res]
        res = [x[1] for x in res]
        [x.update(kw) for x in res]
        res.sort(key=lambda x: np.median(x['fitness']), reverse=True)
        res = json.dumps(res, sort_keys=True, indent=2)
        if parameters.endswith('.gz'):
            with gzip.open(parameters, 'wb') as fpt:
                try:
                    fpt.write(bytes(res, encoding='utf-8'))
                except TypeError:
                    fpt.write(res)
        else:
            with open(parameters, 'w') as fpt:
                fpt.write(res)

    def main(self):
        self.read_training_set()
        kw = {}
        if self.data.classifier:
            self.data.multiple_outputs = True
        elif self.data.output_dim > 1:
            self.data.multiple_outputs = True
        for k, v in EvoDAG().get_params().items():
            if hasattr(self.data, k) and getattr(self.data, k) is not None:
                kw[k] = getattr(self.data, k)
        self.evolve(kw)


class CommandLineTrain(CommandLine):
    def __init__(self):
        self.Xtest = None
        self.inputs = Inputs()
        self.word2id = self.inputs.word2id
        self.label2id = self.inputs.label2id
        self.parser = argparse.ArgumentParser(description="EvoDAG")
        self.training_set()
        self.parameters()
        self.model()
        self.cores()
        self.ensemble()
        self.test_set()
        self.version()

    def parameters(self):
        cdn = 'File containing a list of parameters explored,\
        the first one being the best'
        g = self.parser.add_mutually_exclusive_group(required=True)
        g.add_argument('-C', '--classifier', dest='classifier',
                       help='The task is classification (default)',
                       default=True,
                       action="store_true")
        g.add_argument('-R', '--regressor', dest='regressor',
                       help='The task is regression',
                       action="store_true")
        g.add_argument('-P', '--parameters',
                       dest='parameters',
                       default=None,
                       type=str,
                       help=cdn)
        self.parser.add_argument('--output-dim',
                                 dest='output_dim',
                                 default=1,
                                 type=int,
                                 help="Output Dimension (default 1) use with multiple-outputs flag")
        dr = self.parser.add_argument
        dr('--kw', dest='kwargs', default=None, type=str,
           help='Parameters in json that overwrite default parameters')

    def model(self):
        cdn = 'File to store EvoDAG model'
        pa = self.parser.add_argument
        pa('-m', '--model',
           dest='model_file',
           type=str,
           help=cdn)
        pa('-j', '--json', dest='json',
           action="store_true",
           help='Whether the inputs are in json format',
           default=False)
        # pa('--min-size', dest='min_size',
        #    type=int, default=1, help='Model min-size')
        pa('-s', '--seed', dest='seed',
           default=-1, type=int, help='Seed')

    def main(self):
        self.read_training_set()
        self.read_test_set()
        model_file = self.get_model_file()
        classifier = False if self.data.regressor else True
        kw = dict(params_fname=self.data.parameters, classifier=classifier)
        if self.data.seed >= 0:
            kw['seed'] = self.data.seed
        if self.data.kwargs is not None:
            _ = json.loads(self.data.kwargs)
            kw.update(_)
        if self.data.ensemble_size == 1:
            evo = model_EvoDAG(**kw).fit(self.X, self.y, test_set=self.Xtest)
            self.model = evo.model
        else:
            self.model = EvoDAGE(n_estimators=self.data.ensemble_size,
                                 n_jobs=self.data.cpu_cores,
                                 tmpdir=model_file + '_dir',
                                 **kw).fit(self.X, self.y,
                                           test_set=self.Xtest)
        with gzip.open(model_file, 'w') as fpt:
            pickle.dump(self.model, fpt)
            pickle.dump(self.word2id, fpt)
            pickle.dump(self.label2id, fpt)


class CommandLinePredict(CommandLine):
    def __init__(self):
        self.Xtest = None
        self.inputs = Inputs()
        self.word2id = self.inputs.word2id
        self.label2id = self.inputs.label2id
        self.parser = argparse.ArgumentParser(description="EvoDAG")
        self.model()
        self.test_set()
        self.output_file()
        self.raw_outputs()
        self.cores()
        self.version()

    def test_set(self):
        cdn = 'File containing the test set on csv.'
        self.parser.add_argument('test_set',
                                 default=None,
                                 help=cdn)

    def raw_outputs(self):
        cdn = 'Raw decision function.'
        self.parser.add_argument('--raw-outputs',
                                 default=False,
                                 dest='raw_outputs',
                                 action='store_true',
                                 help=cdn)

    def model(self):
        cdn = 'EvoDAG model'
        pa = self.parser.add_argument
        pa('-m', '--model',
           dest='model_file',
           type=str,
           help=cdn)
        pa('-j', '--json', dest='json',
           action="store_true",
           help='Whether the inputs are in json format',
           default=False)
        pa('--decision-function', dest='decision_function', default=False,
           action='store_true',
           help='Outputs the decision functions instead of the class')

    def main(self):
        model_file = self.get_model_file()
        with gzip.open(model_file, 'r') as fpt:
            m = pickle.load(fpt)
            self.inputs.word2id.update(pickle.load(fpt))
            self.inputs.label2id.update(pickle.load(fpt))
        self.read_test_set()
        self.data.classifier = m.classifier
        if self.data.raw_outputs:
            hy = m.raw_outputs(self.Xtest,
                               cpu_cores=self.data.cpu_cores)
            if hy.ndim == 3:
                hy.shape = (hy.shape[1] * hy.shape[0], hy.shape[-1])
            hy = "\n".join([",".join([str(i) for i in x]) for x in hy.T])
        elif self.data.decision_function:
            hy = m.decision_function(self.Xtest, cpu_cores=self.data.cpu_cores)
            if isinstance(hy, SparseArray):
                hy = tonparray(hy)
                hy = "\n".join(map(str, hy))
            else:
                hy = np.array([tonparray(x) for x in hy]).T
                hy = "\n".join([",".join([str(i) for i in x]) for x in hy])
        else:
            hy = self.id2label(m.predict(self.Xtest, cpu_cores=self.data.cpu_cores))
            hy = "\n".join(map(str, hy))
        with open(self.get_output_file(), 'w') as fpt:
            fpt.write(hy)


class CommandLineUtils(CommandLine):
    def __init__(self):
        self.Xtest = None
        self.inputs = Inputs()
        self.word2id = self.inputs.word2id
        self.label2id = self.inputs.label2id
        self.parser = argparse.ArgumentParser(description="EvoDAG")
        self.model()
        self.graphviz()
        self.params_stats()
        self.output_file()
        self.fitness()
        self.size()
        self.height()
        self.remove_terminals()
        self.used_inputs_number()
        self.create_ensemble_params()
        self.cores()
        self.version()

    def create_ensemble_params(self):
        self.parser.add_argument('--create-ensemble',
                                 help='Models to ensemble', dest='ensemble',
                                 default=False, action='store_true')
        self.parser.add_argument('--best-params-file',
                                 help='Search for the best configuration in a given directory', dest='best_params_file', default=False, action='store_true')
        self.parser.add_argument('-n', '--ensemble-size',
                                 help='Ensemble size (default: select all models)',
                                 dest='ensemble_size',
                                 default=-1,
                                 type=int)

    def used_inputs_number(self):
        self.parser.add_argument('--used-inputs-number',
                                 help='Number of inputs used',
                                 dest='used_inputs_number',
                                 default=False, action='store_true')

    def remove_terminals(self):
        self.parser.add_argument('--remove-terminals',
                                 help='Do not display terminals',
                                 dest='remove_terminals',
                                 default=False, action='store_true')

    def height(self):
        self.parser.add_argument('--height',
                                 help='Model height',
                                 dest='height',
                                 default=False, action='store_true')

    def size(self):
        self.parser.add_argument('--size',
                                 help='Model size',
                                 dest='size',
                                 default=False, action='store_true')

    def fitness(self):
        self.parser.add_argument('--fitness',
                                 help='Fitness in the validation set',
                                 dest='fitness',
                                 default=False, action='store_true')

    def graphviz(self):
        self.parser.add_argument('-G', '--graphviz',
                                 help='Plot the model using dot language',
                                 dest='graphviz',
                                 default=False, action='store_true')

    def params_stats(self):
        self.parser.add_argument('-P', '--params-stats',
                                 help='Parameters statistics',
                                 dest='params_stats',
                                 default=False, action='store_true')

    def output_file(self):
        self.parser.add_argument('-o', '--output-file',
                                 help='File / directory to store the result(s)',
                                 dest='output_file',
                                 default=None,
                                 type=str)

    def model(self):
        cdn = 'File containing the model/params.'
        self.parser.add_argument('model_file',
                                 default=None,
                                 type=str,
                                 help=cdn)

    def read_params(self, parameters):
        if parameters.endswith('.gz'):
            with gzip.open(parameters, 'rb') as fpt:
                try:
                    res = fpt.read()
                    return json.loads(str(res, encoding='utf-8'))
                except TypeError:
                    return json.loads(res)
        else:
            with open(parameters, 'r') as fpt:
                return json.loads(fpt.read())

    def create_ensemble(self, model_file):
        from glob import glob
        models = []
        flag = False
        for fname in model_file.split(' '):
            for k in tqdm(glob(fname)):
                try:
                    with gzip.open(k, 'r') as fpt:
                        models.append(pickle.load(fpt))
                        self.inputs.word2id.update(pickle.load(fpt))
                        self.inputs.label2id.update(pickle.load(fpt))
                except EOFError:
                    flag = True
                    os.unlink(k)
        if flag:
            raise RuntimeError('Unable to read models')
        models.sort(key=lambda x: x.fitness_vs, reverse=True)
        if self.data.ensemble_size > 0:
            models = models[:self.data.ensemble_size]
        self.model = Ensemble(models)
        model_file = self.data.output_file
        with gzip.open(model_file, 'w') as fpt:
            pickle.dump(self.model, fpt)
            pickle.dump(self.word2id, fpt)
            pickle.dump(self.label2id, fpt)

    def get_best_params(self, model_file):
        from glob import glob
        h = {}
        args = glob('%s/*.model' % model_file)
        if self.data.cpu_cores == 1:
            res = [get_model_fitness(x) for x in tqdm(args)]
        else:
            p = Pool(self.data.cpu_cores)
            res = [x for x in tqdm(p.imap_unordered(get_model_fitness, args),
                                   total=len(args))]
            p.close()
        for m, fit in res:
            basename = (m.split(os.path.join(model_file, ''))[1]).split('_')[:1]
            fname = "_".join(basename)
            try:
                h[fname].append(fit)
            except KeyError:
                h[fname] = [fit]
        b = max(h.items(), key=lambda x: np.median(x[1]))
        self.best_params = b[0] + '_params.json'

    def main(self):
        def most_common(K, a):
            try:
                str_type = unicode
            except NameError:
                str_type = str

            l = a.most_common()
            if len(l):
                if len(PARAMS[K]) <= 2:
                    return l[0]
                elif isinstance(l[0][0], str_type):
                    return l[0]
                else:
                    num = np.sum([x * y for x, y in a.items()])
                    den = float(np.sum([y for y in a.values()]))
                    return num / den
            return ""

        model_file = self.get_model_file()
        if self.data.graphviz:
            with gzip.open(model_file, 'r') as fpt:
                m = pickle.load(fpt)
                self.inputs.word2id.update(pickle.load(fpt))
                self.inputs.label2id.update(pickle.load(fpt))
            remove_terminals = self.data.remove_terminals
            if remove_terminals:
                m.graphviz(self.data.output_file, terminals=False)
            else:
                m.graphviz(self.data.output_file)
        elif self.data.params_stats:
            params = {k: collections.Counter() for k in PARAMS.keys()}
            stats = self.read_params(model_file)
            for l in stats:
                for k, v in l.items():
                    if k not in params:
                        continue
                    params[k][v] += 1
            with open(self.data.output_file, 'w') as fpt:
                fpt.write(json.dumps({k: most_common(k, v) for k, v
                                      in params.items()}, sort_keys=True, indent=2))
        elif self.data.fitness:
            with gzip.open(model_file, 'r') as fpt:
                m = pickle.load(fpt)
                self.inputs.word2id.update(pickle.load(fpt))
                self.inputs.label2id.update(pickle.load(fpt))
            print("Median fitness: %0.4f" % (m.fitness_vs * -1))
        elif self.data.size:
            with gzip.open(model_file, 'r') as fpt:
                m = pickle.load(fpt)
                self.inputs.word2id.update(pickle.load(fpt))
                self.inputs.label2id.update(pickle.load(fpt))
            print("Size: %s" % m.size)
        elif self.data.height:
            with gzip.open(model_file, 'r') as fpt:
                m = pickle.load(fpt)
                self.inputs.word2id.update(pickle.load(fpt))
                self.inputs.label2id.update(pickle.load(fpt))
            print("Height: %s" % m.height)
        elif self.data.used_inputs_number:
            with gzip.open(model_file, 'r') as fpt:
                m = pickle.load(fpt)
                self.word2id = pickle.load(fpt)
                self.label2id = pickle.load(fpt)
            inputs = m.inputs()
            print("Used inputs number", len(inputs))
        elif self.data.ensemble:
            self.create_ensemble(model_file)
        elif self.data.best_params_file:
            self.get_best_params(model_file)
            print(self.best_params)


def params(output=False):
    "EvoDAG-params command line"
    c = CommandLineParams()
    c.parse_args()
    if output:
        return c


def train(output=False):
    "EvoDAG-params command line"
    c = CommandLineTrain()
    c.parse_args()
    if output:
        return c


def predict():
    "EvoDAG-params command line"
    c = CommandLinePredict()
    c.parse_args()


def utils(output=False):
    "EvoDAG-utils command line"
    c = CommandLineUtils()
    c.parse_args()
    if output:
        return c



# Copyright 2017 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nose.tools import assert_almost_equals


def create_problem_node2(nargs=4, seed=0):
    from EvoDAG import RootGP
    from test_root import X, cl
    import numpy as np
    gp = RootGP(generations=1, popsize=nargs,
                multiple_outputs=True, seed=seed)
    X1 = np.concatenate((X, np.atleast_2d(np.zeros(X.shape[0])).T), axis=1)
    for i in range(10, 20):
        X1[i, -1] = 1
    gp.X = X1
    gp.Xtest = X1
    y = cl.copy()
    gp.nclasses(y)
    gp.y = y
    return gp, [gp.X[x] for x in range(nargs)]


def test_mean_std_pk():
    import numpy as np
    from EvoDAG.naive_bayes import NaiveBayes
    gp, args = create_problem_node2(nargs=5, seed=0)
    mask = np.array(gp._mask_ts.sign().full_array(), dtype=np.bool)
    klass = np.array(gp._y_klass.full_array())[mask]
    unique_klass = np.unique(klass)
    naive_bayes = NaiveBayes(mask=gp._mask_ts, klass=gp._y_klass,
                             nclass=gp._labels.shape[0])
    for v in args:
        var = np.array(v._eval_tr.full_array())[mask]
        mean = [np.mean(var[k == klass]) for k in unique_klass]
        std2 = [np.var(var[k == klass]) for k in unique_klass]
        p_klass = [(k == klass).mean() for k in unique_klass]
        _mean, _std2, _klass = naive_bayes.coef(v._eval_tr)
        for a, b in zip(mean, _mean):
            assert_almost_equals(a, b)
        for a, b in zip(std2, _std2):
            assert_almost_equals(a, b)
        for a, b in zip(p_klass, _klass):
            assert_almost_equals(a, b)


def test_MN_Nc_pk():
    import numpy as np
    from EvoDAG.naive_bayes import NaiveBayes
    gp, args = create_problem_node2(nargs=5, seed=0)
    mask = np.array(gp._mask_ts.sign().full_array(), dtype=np.bool)
    klass = np.array(gp._y_klass.full_array())[mask]
    unique_klass = np.unique(klass)
    naive_bayes = NaiveBayes(mask=gp._mask_ts, klass=gp._y_klass,
                             nclass=gp._labels.shape[0])
    for v in args:
        var = np.array(v._eval_tr.full_array())[mask]
        Nc = [np.sum([x for x in var[k == klass] if x > 0]) for k in
              unique_klass]
        if len(Nc) > len([x for x in Nc if x > 0]):
            continue
        N = np.sum(Nc).reshape(-1, 1)
        w = np.log((Nc / N)[0])
        p_klass = np.log([(k == klass).mean() for k in unique_klass])
        _w, _klass = naive_bayes.coef_MN(v._eval_tr)
        for a, b in zip(w, _w):
            assert_almost_equals(a, b)
        for a, b in zip(p_klass, _klass):
            assert_almost_equals(a, b)



# Copyright 2017 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from SparseArray.sparse_array cimport SparseArray
from cpython.list cimport PyList_GET_SIZE, PyList_GET_ITEM, PyList_New, PyList_Append
from cpython cimport array
from cpython cimport list
from libc cimport math
import random
cimport cython


cpdef double fitness_SSE(list _ytr, list _hy, SparseArray _mask):
    cdef SparseArray ytr, hy, mask
    cdef double res = 0
    cdef Py_ssize_t i
    for i in range(PyList_GET_SIZE(_ytr)):
        ytr = <SparseArray> PyList_GET_ITEM(_ytr, i)
        hy = <SparseArray> PyList_GET_ITEM(_hy, i)
        res += -ytr.SSE(hy.mul(_mask))
    return res / len(_ytr)


cpdef double fitness_SAE(list _ytr, list _hy, list _mask):
    cdef SparseArray ytr, hy, mask
    cdef double res = 0
    for ytr, hy, mask in zip(_ytr, _hy, _mask):
        res += -ytr.SAE(hy.mul(mask))
    return res / len(_ytr)


cpdef bint naive_bayes_isfinite(list coef, int nclass):
    cdef Py_ssize_t i, j=0
    cdef array.array c
    cdef double *c_value
    for j in range(1, 3):
        c = <array.array> PyList_GET_ITEM(coef, j)
        c_value = c.data.as_doubles
        for i in range(nclass):
            if c_value[i] == 0:
                return False
    return True


cpdef bint naive_bayes_isfinite_MN(list coef, int nclass):
    cdef Py_ssize_t i, j=0
    cdef array.array c
    cdef double *c_value
    for j in range(2):
        c = <array.array> PyList_GET_ITEM(coef, j)
        c_value = c.data.as_doubles
        for i in range(nclass):
            if not math.isfinite(c_value[i]):
                return False
    return True


cpdef list naive_bayes_mean_std2(SparseArray var, array.array klass, array.array mask, int nclass):
    cdef array.array mean_num, mean_den, std_num
    cdef unsigned int *a = var.index.data.as_uints, index
    cdef unsigned int *klass_value = klass.data.as_uints
    cdef double *a_value = var.data.data.as_doubles, tmp
    cdef double *mean_num_value, *mean_den_value, *std_num_value
    cdef unsigned int *mask_value = mask.data.as_uints
    mean_num = array.clone(var.data, nclass, zero=True)
    mean_num_value = mean_num.data.as_doubles
    std_num = array.clone(var.data, nclass, zero=True)
    std_num_value = std_num.data.as_doubles
    mean_den = array.clone(var.data, nclass, zero=True)
    mean_den_value = mean_den.data.as_doubles
    cdef Py_ssize_t i, k=0
    cdef unsigned int var_end = var.non_zero
    if var_end == 0:
        return [mean_num, std_num, mean_den]
    for i in range(len(mask)):
        index = mask_value[i]
        for k in range(k, var_end):
            if a[k] >= index:
                break
        if a[k] == index:
            mean_num_value[klass_value[index]] += a_value[k]
            mean_den_value[klass_value[index]] += 1
        else:
            mean_den_value[klass_value[index]] += 1
    for i in range(nclass):
        mean_num_value[i] = mean_num_value[i] / mean_den_value[i]
    k = 0
    for i in range(len(mask)):
        index = mask_value[i]
        for k in range(k, var_end):
            if a[k] >= index:
                break
        if a[k] == index:
            tmp = a_value[k] - mean_num[klass_value[index]]
        else:
            tmp = mean_num[klass_value[index]]
        tmp *= tmp
        std_num_value[klass_value[index]] += tmp
    for i in range(nclass):
        std_num_value[i] = std_num_value[i] / mean_den_value[i]
    tmp = len(mask)
    for i in range(nclass):
        mean_den_value[i] = mean_den_value[i] / tmp
    return [mean_num, std_num, mean_den]


@cython.cdivision(True)
cpdef list naive_bayes(list var, list coef, unsigned int nclass):
    cdef Py_ssize_t i, j
    cdef list l = <list> PyList_GET_ITEM(coef, 0)
    cdef array.array p_klass = <array.array> PyList_GET_ITEM(l, 2), s_klass, m_klass
    cdef double *p_klass_value = p_klass.data.as_doubles, a, *s_klass_value, *m_klass_value
    cdef double c
    cdef SparseArray b = <SparseArray> PyList_GET_ITEM(var, 0), v, tmp
    cdef unsigned int _len = len(b), nvar = len(var)
    cdef list res = []
    for i in range(nclass):
        a = math.log(p_klass_value[i])
        b = b.mul2(0)
        c = 0
        for j in range(nvar):
            v = <SparseArray> PyList_GET_ITEM(var, j)
            l = <list> PyList_GET_ITEM(coef, j)
            m_klass = <array.array> PyList_GET_ITEM(l, 0)
            m_klass_value = m_klass.data.as_doubles
            s_klass = <array.array> PyList_GET_ITEM(l, 1)
            s_klass_value = s_klass.data.as_doubles
            c += math.log(2 * math.pi * s_klass_value[i])
            b = b.add(((v.add2(-m_klass_value[i])).sq()).mul2(1 / s_klass_value[i]))
        b = b.mul2(-0.5)
        b = b.add2(a - (0.5 * c))
        PyList_Append(res, b)
    return res


@cython.cdivision(True)
cpdef list naive_bayes_Nc(SparseArray var, array.array klass, array.array mask, int nclass):
    cdef array.array mean_num, mean_den
    cdef unsigned int *a = var.index.data.as_uints, index
    cdef unsigned int *klass_value = klass.data.as_uints
    cdef double *a_value = var.data.data.as_doubles, tmp
    cdef double *mean_num_value, *mean_den_value, N=0
    cdef unsigned int *mask_value = mask.data.as_uints
    mean_num = array.clone(var.data, nclass, zero=True)
    mean_num_value = mean_num.data.as_doubles
    mean_den = array.clone(var.data, nclass, zero=True)
    mean_den_value = mean_den.data.as_doubles
    cdef Py_ssize_t i, k=0
    cdef unsigned int var_end = var.non_zero
    if var_end == 0:
        return [mean_num, mean_den]
    for i in range(len(mask)):
        index = mask_value[i]
        for k in range(k, var_end):
            if a[k] >= index:
                break
        if a[k] == index and a_value[k] > 0:
            mean_num_value[klass_value[index]] += a_value[k]
            N += a_value[k]
            mean_den_value[klass_value[index]] += 1
        else:
            mean_den_value[klass_value[index]] += 1
    for i in range(nclass):
        mean_num_value[i] = math.log(mean_num_value[i] / N)
    tmp = len(mask)
    for i in range(nclass):
        mean_den_value[i] = math.log(mean_den_value[i] / tmp)
    return [mean_num, mean_den]


@cython.cdivision(True)
cpdef list naive_bayes_MN(list var, list coef, unsigned int nclass):
    cdef Py_ssize_t i, j, k
    cdef list l = <list> PyList_GET_ITEM(coef, 0)
    cdef array.array p_klass = <array.array> PyList_GET_ITEM(l, 1), m_klass
    cdef double *p_klass_value = p_klass.data.as_doubles, *m_klass_value
    cdef SparseArray b = <SparseArray> PyList_GET_ITEM(var, 0), v
    cdef unsigned int nvar = len(var)
    cdef list res = []
    for i in range(nclass):
        b = b.mul2(0)
        for j in range(nvar):
            v = <SparseArray> PyList_GET_ITEM(var, j)
            l = <list> PyList_GET_ITEM(coef, j)
            m_klass = <array.array> PyList_GET_ITEM(l, 0)
            m_klass_value = m_klass.data.as_doubles
            b = b.add(v.mul2(m_klass_value[i]))
        # b = b.add2(p_klass_value[i])
        b = b.add(SparseArray.constant(p_klass_value[i], b.index, b._len))
        m_klass_value = b.data.data.as_doubles
        for k in range(b.non_zero):
            m_klass_value[k] = math.exp(m_klass_value[k])
        PyList_Append(res, b)
    return res


cdef class SelectNumbers:
    cdef array.array data
    cdef public Py_ssize_t pos
    cdef public Py_ssize_t size
    def __cinit__(self, list lst):
        random.shuffle(lst)
        self.data = array.array('I', lst)
        self.size = PyList_GET_SIZE(lst)
        self.pos = 0

    cpdef list get(self, Py_ssize_t k):
        cdef Py_ssize_t end = self.pos + k, i, pos = self.pos
        cdef list res = PyList_New(0)
        cdef array.array data = self.data
        if end > self.size:
            end = self.size
        self.pos = end
        for i in range(pos, end):
            PyList_Append(res, data[i])
        return res

    cpdef int get_one(self):
        cdef pos = self.pos
        self.pos += 1
        return self.data[pos]
    
    cpdef bint empty(self):
        return self.pos == self.size


cdef class Score:
    cdef public array.array precision
    cdef public array.array precision_den
    cdef public array.array recall
    cdef public array.array recall_den
    cdef public array.array precision2
    cdef public array.array precision2_den
    cdef public array.array recall2
    cdef public array.array recall2_den
    cdef public double _accuracy
    cdef public double _accuracy2
    cdef unsigned int y_pos
    cdef unsigned int hy_pos
    cdef unsigned int y_size
    cdef unsigned int hy_size
    cdef unsigned int nclasses

    def __cinit__(self, unsigned int nclasses):
        cdef  array.array mask = array.array('d', [])
        self.precision = array.clone(mask, nclasses, zero=False)
        self.recall = array.clone(mask, nclasses, zero=False)
        self.precision2 = array.clone(mask, nclasses, zero=False)
        self.recall2 = array.clone(mask, nclasses, zero=False)
        mask = array.array('I', [])
        self.precision_den = array.clone(mask, nclasses, zero=False)
        self.recall_den = array.clone(mask, nclasses, zero=False)
        self.precision2_den = array.clone(mask, nclasses, zero=False)
        self.recall2_den = array.clone(mask, nclasses, zero=False)
        self.nclasses = nclasses

    cdef _cleanI(self, array.array d):
        cdef Py_ssize_t i = 0
        cdef unsigned int *a = d.data.as_uints
        for i in range(self.nclasses):
            a[i] = 0

    cdef _cleand(self, array.array d):
        cdef Py_ssize_t i = 0
        cdef double *a = d.data.as_doubles
        for i in range(self.nclasses):
            a[i] = 0
            
    cdef clean(self):
        self._cleand(self.precision)
        self._cleanI(self.precision_den)
        self._cleand(self.recall)
        self._cleanI(self.recall_den)
        self._cleand(self.precision2)
        self._cleanI(self.precision2_den)
        self._cleand(self.recall2)
        self._cleanI(self.recall2_den)
        self.y_pos = 0
        self.hy_pos = 0
        self._accuracy = 0
        self._accuracy2 = 0

    cdef double get_klass(self, double *data, unsigned int *index, unsigned int i):
        cdef double res = 0
        if self.y_pos >= self.y_size:
            return res
        if index[self.y_pos] == i:
            res = data[self.y_pos]
            self.y_pos += 1
        return res

    cdef double get_predicted(self, double *data, unsigned int *index, unsigned int i):
        cdef double res = 0
        if self.hy_pos >= self.hy_size:
            return res
        if index[self.hy_pos] == i:
            res = data[self.hy_pos]
            self.hy_pos += 1
        return res

    def g_recall_precision(self, Py_ssize_t i, SparseArray y,
                           SparseArray hy, array.array index):
        self.count(y, hy, index)
        self.precision_recall()
        cdef double *precision = self.precision.data.as_doubles
        cdef double *recall = self.recall.data.as_doubles
        cdef double *precision2 = self.precision2.data.as_doubles
        cdef double *recall2 = self.recall2.data.as_doubles
        cdef double f1 = 0, f12 = 0
        f1 = precision[i] * recall[i]
        f12 = precision2[i] * recall2[i]
        return f1, f12
    
    @cython.cdivision(True)    
    def F1(self, Py_ssize_t i, SparseArray y, SparseArray hy, array.array index):
        self.count(y, hy, index)
        self.precision_recall()
        cdef double *precision = self.precision.data.as_doubles
        cdef double *recall = self.recall.data.as_doubles
        cdef double *precision2 = self.precision2.data.as_doubles
        cdef double *recall2 = self.recall2.data.as_doubles
        cdef double f1 = 0, f12 = 0, den
        den = precision[i] + recall[i]
        if den > 0:
            f1 = (2 * precision[i] * recall[i]) / den
        den = precision2[i] + recall2[i]
        if den > 0:
            f12 = (2 * precision2[i] * recall2[i]) / den
        return f1, f12

    @cython.cdivision(True)    
    def macroRecallF1(self, SparseArray y, SparseArray hy, array.array index):
        self.count(y, hy, index)
        self.precision_recall()
        cdef double *recall = self.recall.data.as_doubles
        cdef double *precision2 = self.precision2.data.as_doubles
        cdef double *recall2 = self.recall2.data.as_doubles
        cdef double f1 = 0, f12 = 0, den
        cdef Py_ssize_t i = 0
        for i in range(self.nclasses):
            f1 += recall[i]
            den = precision2[i] + recall2[i]
            if den > 0:
                f12 += (2 * precision2[i] * recall2[i]) / den
        f1 = f1 / self.nclasses
        f12 = f12 / self.nclasses
        return f1, f12

    @cython.cdivision(True)    
    def a_recall(self, SparseArray y, SparseArray hy, array.array index):
        self.count(y, hy, index)
        self.precision_recall()
        cdef double *recall = self.recall.data.as_doubles
        cdef double *recall2 = self.recall2.data.as_doubles
        cdef double f1 = 0, f12 = 0, den
        cdef Py_ssize_t i = 0
        for i in range(self.nclasses):
            f1 += recall[i]
            f12 += recall2[i]
        f1 = f1 / self.nclasses
        f12 = f12 / self.nclasses
        return f1, f12

    @cython.cdivision(True)    
    def a_precision(self, SparseArray y, SparseArray hy, array.array index):
        self.count(y, hy, index)
        self.precision_recall()
        cdef double *recall = self.precision.data.as_doubles
        cdef double *recall2 = self.precision2.data.as_doubles
        cdef double f1 = 0, f12 = 0, den
        cdef Py_ssize_t i = 0
        for i in range(self.nclasses):
            f1 += recall[i]
            f12 += recall2[i]
        f1 = f1 / self.nclasses
        f12 = f12 / self.nclasses
        return f1, f12

    def g_precision(self, SparseArray y, SparseArray hy, array.array index):
        self.count(y, hy, index)
        self.precision_recall()
        cdef double *recall = self.precision.data.as_doubles
        cdef double *recall2 = self.precision2.data.as_doubles
        cdef double f1 = 1, f12 = 1
        cdef Py_ssize_t i = 0
        for i in range(self.nclasses):
            f1 *= recall[i]
            f12 *= recall2[i]
        return f1, f12    

    def g_recall(self, SparseArray y, SparseArray hy, array.array index):
        self.count(y, hy, index)
        self.precision_recall()
        cdef double *recall = self.recall.data.as_doubles
        cdef double *recall2 = self.recall2.data.as_doubles
        cdef double f1 = 1, f12 = 1
        cdef Py_ssize_t i = 0
        for i in range(self.nclasses):
            f1 *= recall[i]
            f12 *= recall2[i]
        return f1, f12
    
    @cython.cdivision(True)
    cdef recall2accuracy(self):
        cdef double *recall = self.recall.data.as_doubles
        cdef double *recall2 = self.recall2.data.as_doubles
        cdef unsigned int *recall_den = self.recall_den.data.as_uints
        cdef unsigned int *recall2_den = self.recall2_den.data.as_uints
        cdef unsigned int den = 0, den2 = 0
        cdef Py_ssize_t i
        for i in range(self.nclasses):
            self._accuracy += recall[i]
            self._accuracy2 += recall2[i]
            den += recall_den[i]
            den2 += recall2_den[i]
        self._accuracy = self._accuracy / den
        self._accuracy2 = self._accuracy2 / den2

    def accuracy(self, SparseArray y, SparseArray hy, array.array index):
        self.count(y, hy, index)
        self.recall2accuracy()
        return self._accuracy, self._accuracy2
        
    @cython.cdivision(True)    
    def accDotMacroF1(self, SparseArray y, SparseArray hy, array.array index):
        cdef double *precision = self.precision.data.as_doubles
        cdef double *recall = self.recall.data.as_doubles
        cdef double *precision2 = self.precision2.data.as_doubles
        cdef double *recall2 = self.recall2.data.as_doubles
        cdef double f1 = 0, f12 = 0, den
        cdef Py_ssize_t i = 0
        self.count(y, hy, index)
        self.recall2accuracy()
        self.precision_recall()
        for i in range(self.nclasses):
            den = precision[i] + recall[i]
            if den > 0:
                f1 += (2 * precision[i] * recall[i]) / den
            den = precision2[i] + recall2[i]
            if den > 0:
                f12 += (2 * precision2[i] * recall2[i]) / den
        f1 = f1 / self.nclasses
        f12 = f12 / self.nclasses
        return f1 * self._accuracy, f12 * self._accuracy2

    def g_g_recall_precision(self, SparseArray y, SparseArray hy, array.array index):
        self.count(y, hy, index)
        self.precision_recall()
        cdef double *precision = self.precision.data.as_doubles
        cdef double *recall = self.recall.data.as_doubles
        cdef double *precision2 = self.precision2.data.as_doubles
        cdef double *recall2 = self.recall2.data.as_doubles
        cdef double f1 = 1, f12 = 1
        cdef Py_ssize_t i = 0
        for i in range(self.nclasses):
            f1 *= precision[i] * recall[i]
            f12 *= precision2[i] * recall2[i]
        return f1, f12
    
    def g_F1(self, SparseArray y, SparseArray hy, array.array index):
        self.count(y, hy, index)
        self.precision_recall()
        cdef double *precision = self.precision.data.as_doubles
        cdef double *recall = self.recall.data.as_doubles
        cdef double *precision2 = self.precision2.data.as_doubles
        cdef double *recall2 = self.recall2.data.as_doubles
        cdef double f1 = 1, f12 = 1, den
        cdef Py_ssize_t i = 0
        for i in range(self.nclasses):
            den = precision[i] + recall[i]
            if den > 0:
                f1 *= (2 * precision[i] * recall[i]) / den
            else:
                f1 = 0
            den = precision2[i] + recall2[i]
            if den > 0:
                f12 *= (2 * precision2[i] * recall2[i]) / den
            else:
                f12 = 0
        return f1, f12
    
    @cython.cdivision(True)    
    def a_F1(self, SparseArray y, SparseArray hy, array.array index):
        self.count(y, hy, index)
        self.precision_recall()
        cdef double *precision = self.precision.data.as_doubles
        cdef double *recall = self.recall.data.as_doubles
        cdef double *precision2 = self.precision2.data.as_doubles
        cdef double *recall2 = self.recall2.data.as_doubles
        cdef double f1 = 0, f12 = 0, den
        cdef Py_ssize_t i = 0
        for i in range(self.nclasses):
            den = precision[i] + recall[i]
            if den > 0:
                f1 += (2 * precision[i] * recall[i]) / den
            den = precision2[i] + recall2[i]
            if den > 0:
                f12 += (2 * precision2[i] * recall2[i]) / den
        f1 = f1 / self.nclasses
        f12 = f12 / self.nclasses
        return f1, f12

    @cython.cdivision(True)
    cdef precision_recall(self):
        cdef double *precision = self.precision.data.as_doubles
        cdef double *recall = self.recall.data.as_doubles
        cdef double *precision2 = self.precision2.data.as_doubles
        cdef double *recall2 = self.recall2.data.as_doubles
        cdef unsigned int *precision_den = self.precision_den.data.as_uints
        cdef unsigned int *recall_den = self.recall_den.data.as_uints
        cdef unsigned int *precision2_den = self.precision2_den.data.as_uints
        cdef unsigned int *recall2_den = self.recall2_den.data.as_uints
        cdef Py_ssize_t i
        for i in range(self.nclasses):
            if precision_den[i] > 0:
                precision[i] = precision[i] / precision_den[i]
            else:
                precision[i] = 0
            if recall_den[i] > 0:
                recall[i] = recall[i] / recall_den[i]
            else:
                recall[i] = 0
            if precision2_den[i] > 0:
                precision2[i] = precision2[i] / precision2_den[i]
            else:
                precision2[i] = 0
            if recall2_den[i] > 0:
                recall2[i] = recall2[i] / recall2_den[i]
            else:
                recall2[i] = 0
        
    @cython.cdivision(True)
    cdef count(self, SparseArray y, SparseArray hy, array.array index):
        self.clean()
        cdef unsigned int *index_value = index.data.as_uints
        cdef bint flag
        cdef unsigned int i = 0, size = y._len
        cdef unsigned int k = 0, end = len(index)
        cdef double *y_data = y.data.data.as_doubles
        cdef unsigned int *y_index = y.index.data.as_uints
        cdef double *hy_data = hy.data.data.as_doubles
        cdef unsigned int *hy_index = hy.index.data.as_uints
        cdef double *precision = self.precision.data.as_doubles
        cdef double *recall = self.recall.data.as_doubles
        cdef double *precision2 = self.precision2.data.as_doubles
        cdef double *recall2 = self.recall2.data.as_doubles
        cdef unsigned int *precision_den = self.precision_den.data.as_uints
        cdef unsigned int *recall_den = self.recall_den.data.as_uints
        cdef unsigned int *precision2_den = self.precision2_den.data.as_uints
        cdef unsigned int *recall2_den = self.recall2_den.data.as_uints
        cdef int _y, _hy
        self.y_size = y.non_zero
        self.hy_size = hy.non_zero
        for i in range(size):
            _y = <int> self.get_klass(y_data, y_index, i)
            _hy = <int> self.get_predicted(hy_data, hy_index, i)
            flag = False
            if k < end:
                if index_value[k] == i:
                    flag = True
                    k += 1
            if flag:
                precision_den[_hy] += 1
                recall_den[_y] += 1
                if _y == _hy:
                    precision[_y] += 1
                    recall[_y] += 1
            else:
                precision2_den[_hy] += 1
                recall2_den[_y] += 1
                if _y == _hy:
                    precision2[_y] += 1
                    recall2[_y] += 1



# Copyright 2013 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
from SparseArray import SparseArray
import json
import os
import gzip


def line_iterator(filename):
    if filename.endswith(".gz"):
        f = gzip.GzipFile(filename)
    else:
        try:
            f = open(filename, encoding='utf8')
        except TypeError:
            f = open(filename)
    while True:
        line = f.readline()
        # Test the type of the line and encode it if neccesary...
        if type(line) is bytes:
            try:
                line = str(line, encoding='utf8')
            except TypeError:
                line = str(line)
        # If the line is empty, we are done...
        if len(line) == 0:
            break
        line = line.strip()
        # If line is empty, jump to next...
        if len(line) == 0:
            continue
        yield line
    # Close the file...
    f.close()


def json_iterator(filename):
    for line in line_iterator(filename):
        yield json.loads(line)


def tonparray(a):
    return np.array(a.full_array())


def BER(y, yh):
    u = np.unique(y)
    b = 0
    for cl in u:
        m = y == cl
        b += (~(y[m] == yh[m])).sum() / float(m.sum())
    return (b / float(u.shape[0])) * 100.


def RSE(x, y):
    return ((x - y)**2).sum() / ((x - x.mean())**2).sum()


params_fname = os.path.join(os.path.dirname(__file__), 'conf', 'parameter_values.json')
with open(params_fname, 'r') as fpt:
    PARAMS = json.loads(fpt.read())


class Inputs(object):
    def __init__(self):
        self._word2id = {}
        self._label2id = {}

    @property
    def word2id(self):
        return self._word2id

    @property
    def label2id(self):
        return self._label2id

    @word2id.setter
    def word2id(self, s):
        self._word2id = s

    @label2id.setter
    def label2id(self, s):
        self._label2id = s

    @staticmethod
    def _num_terms(a):
        if 'num_terms' in a:
            num_terms = a['num_terms']
        else:
            num_terms = len(a)
            if 'klass' in a:
                num_terms -= 1
        return num_terms

    def convert(self, x):
        try:
            return float(x)
        except ValueError:
            if x not in self.word2id:
                self.word2id[x] = len(self.word2id)
            return self.word2id[x]

    def convert_label(self, x):
        try:
            x = float(x)
            if np.isfinite(x):
                return x
            x = str(x)
        except ValueError:
            pass
        if x not in self.label2id:
            self.label2id[x] = len(self.label2id)
        return self.label2id[x]

    def _read_csv(self, fname):
        X = []
        for i in line_iterator(fname):
            x = i.rstrip().lstrip()
            if len(x):
                X.append([i for i in x.split(',')])
        return X

    def read_csv(self, fname, dim):
        X = []
        y = []
        d = self._read_csv(fname)
        if dim > 1:
            for x in d:
                X.append([self.convert(i) for i in x[:-dim]])
                y.append(x[-dim:])
            X = np.array(X)
            y = [SparseArray.fromlist([float(x[i]) for x in y]) for i in range(dim)]
        elif dim == 1:
            for x in d:
                X.append([self.convert(i) for i in x[:-1]])
                y.append(self.convert_label(x[-1]))
            X = np.array(X)
            y = np.array(y)
        else:
            X, y = np.array([[self.convert(i) for i in x] for x in d]), None
        return X, y

    def read_data_json(self, fname, iterable=None):
        X = None
        y = []
        dependent = os.getenv('KLASS')
        if dependent is None:
            dependent = 'klass'
        if iterable is None:
            iterable = json_iterator(fname)
        for row, a in enumerate(iterable):
            if 'vec' in a:
                return self.read_data_json_vec(fname)
            if X is None:
                X = [list() for i in range(self._num_terms(a))]
            for k, v in a.items():
                try:
                    k = int(k)
                    X[k].append((row, self.convert(v)))
                except ValueError:
                    if k == dependent:
                        y.append(self.convert_label(v))
        num_rows = row + 1
        X = [SparseArray.index_data(x, num_rows) for x in X]
        if len(y) == 0:
            y = None
        else:
            y = np.array(y)
        return X, y

    def read_data_json_vec(self, fname, iterable=None):
        X = None
        y = []
        dependent = os.getenv('KLASS')
        if dependent is None:
            dependent = 'klass'
        if iterable is None:
            iterable = json_iterator(fname)
        for row, a in enumerate(iterable):
            vec = a['vec']
            try:
                vecsize = a['vecsize']
            except KeyError:
                vecsize = len(vec)
                vec = enumerate(vec)
            if X is None:
                X = [list() for i in range(vecsize)]
            for k, v in vec:
                k = int(k)
                X[k].append((row, self.convert(v)))
            try:
                y.append(self.convert_label(a[dependent]))
            except KeyError:
                pass
        num_rows = row + 1
        X = [SparseArray.index_data(x, num_rows) for x in X]
        if len(y) == 0:
            y = None
        else:
            y = np.array(y)
        return X, y


class RandomParameterSearch(object):
    def __init__(self, params=PARAMS,
                 npoints=1468,
                 training_size=5000,
                 seed=0):
        self._training_size = training_size
        self.popsize_constraint(params)
        self._params = sorted(params.items())
        assert len(self._params)
        self._params.reverse()
        self._len = None
        self._npoints = npoints
        self._seed = seed
        self.fix_early_popsize()

    def popsize_constraint(self, params):
        try:
            params['popsize'] = [x for x in params['popsize'] if x <= self._training_size]
        except KeyError:
            pass

    def fix_early_popsize(self):
        try:
            popsize = [x for x in self._params if x[0] == 'popsize'][0]
            if len(popsize[1]) == 0:
                popsize[1].append(self._training_size)
        except IndexError:
            pass
        try:
            early = [x for x in self._params if x[0] == 'early_stopping_rounds'][0]
            early_min = min(early[1])
            if early_min > self._training_size:
                early[1].append(self._training_size)
        except IndexError:
            pass

    def __len__(self):
        if self._len is None:
            _ = np.array([len(x[1]) for x in self._params], dtype=np.uint)
            _ = np.product(_)
            assert _ >= 0
            self._len = _
        return self._len

    def __getitem__(self, key):
        res = {}
        lens = [len(x[1]) for x in self._params]
        for l, k_v in zip(lens, self._params):
            k, v = k_v
            key, residual = divmod(key, l)
            res[k] = v[int(residual)]
        return res

    def constraints(self, k):
        try:
            if k['population_class'] == 'Generational' and\
               k['early_stopping_rounds'] < k['popsize']:
                return False
            if k['early_stopping_rounds'] > self._training_size:
                return False
        except KeyError:
            return True
        return True

    def __iter__(self):
        np.random.seed(self._seed)
        m = {}
        _len = self.__len__()
        npoints = self._npoints if _len > self._npoints else _len
        while npoints:
            k = np.round(np.random.uniform(0, _len)).astype(np.uint)
            if len(m) == _len:
                return
            while k in m:
                k = np.random.randint(_len)
            m[k] = 1
            p = self[k]
            if self.constraints(p):
                npoints -= 1
                yield p

    @staticmethod
    def process_params(a):
        from EvoDAG import EvoDAG
        fs_class = {}
        function_set = []
        for x in EvoDAG()._function_set:
            fs_class[x.__name__] = x
        args = {}
        for k, v in a.items():
            if k in fs_class:
                if not isinstance(v, bool):
                    fs_class[k].nargs = v
                if v:
                    function_set.append(fs_class[k])
            else:
                args[k] = v
            fs_evo = EvoDAG()._function_set
            fs_evo = filter(lambda x: x in function_set, fs_evo)
            args['function_set'] = [x for x in fs_evo]
        return args



# Copyright 2017 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .cython_utils import naive_bayes_mean_std2, naive_bayes_Nc
import array


class NaiveBayes(object):
    def __init__(self, mask=None, nclass=None, klass=None):
        self._mask = mask.index
        self._nclass = nclass
        self._klass = array.array('I', [int(x) for x in klass.full_array()])

    def coef(self, var):
        return naive_bayes_mean_std2(var, self._klass, self._mask, self._nclass)

    def coef_MN(self, var):
        return naive_bayes_Nc(var, self._klass, self._mask, self._nclass)





# Copyright 2016 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
cimport numpy as np
from SparseArray.sparse_array cimport SparseArray
from libc cimport math
from libc.float cimport DBL_EPSILON
from cpython cimport array
from cpython.list cimport PyList_GET_SIZE, PyList_GET_ITEM, PyList_SET_ITEM, PyList_Append


cdef bint iszero(double x):
    if x >= DBL_EPSILON or x <= -DBL_EPSILON:
        return False
    return True


cdef double zero_round(double value):
    if iszero(value):
        return 0.0
    return value

        
cdef void swap(list m, Py_ssize_t i):
    cdef double max, comp
    cdef Py_ssize_t pos=i, size=PyList_GET_SIZE(m)
    cdef array.array data, tmp
    # data = m[i]
    data = <array.array> PyList_GET_ITEM(m, i)
    max = math.fabs(data.data.as_doubles[i])
    for j in range(i+1, size):
        # data = m[j]
        data = <array.array> PyList_GET_ITEM(m, j)
        comp = math.fabs(data.data.as_doubles[i])
        if comp > max:
            max = comp
            pos = j
    if pos == i:
        return
    # data = m[i]
    data = <array.array> PyList_GET_ITEM(m, i)
    # m[i] = m[pos]
    tmp = <array.array> PyList_GET_ITEM(m, pos)
    PyList_SET_ITEM(m, i, tmp)
    # m[pos] = data
    PyList_SET_ITEM(m, pos, data)
    

cdef bint gauss_jordan(list m):
    cdef Py_ssize_t i, j, k, size=PyList_GET_SIZE(m)
    cdef array.array data, below
    cdef double c, tmp, c_den
    cdef double *data_values, *below_values
    for i in range(size):
        swap(m, i)
        # data = m[i]
        data = <array.array> PyList_GET_ITEM(m, i)
        data_values = data.data.as_doubles
        c_den = data_values[i]
        if iszero(c_den):
            return True
        for j in range(i+1, size):
            # below = m[j]
            below = <array.array> PyList_GET_ITEM(m, j)
            below_values = below.data.as_doubles
            c = below_values[i] / c_den
            for k in range(i, size+1):
                tmp = below_values[k] - data_values[k] * c
                below_values[k] = tmp
    for i in range(size-1, -1, -1):
        # data = m[i]
        data = <array.array> PyList_GET_ITEM(m, i)
        data_values = data.data.as_doubles
        c = data_values[i]
        if iszero(c):
            return True
        for j in range(0, i):
            # below = m[j]
            below = <array.array> PyList_GET_ITEM(m, j)
            below_values = below.data.as_doubles
            k = size
            tmp = below_values[k] - data_values[k] *\
                  below_values[i] / c
            below_values[k] = zero_round(tmp)
        tmp = data_values[size] / c
        if not math.isfinite(tmp):
            return True
        data_values[size] = zero_round(tmp)
    return False

        
cpdef compute_weight(list r, SparseArray ytr, mask):
    """Returns the weight (w) using OLS of r * w = gp._ytr """
    cdef Py_ssize_t i, j, size=PyList_GET_SIZE(r), k=0
    cdef SparseArray ri, rj
    cdef array.array data = array.array('d'), other, dependent
    cdef list X = [], R = [], var = []
    cdef double tmp, *dependent_value
    dependent = array.clone(data, size, False)
    dependent_value = dependent.data.as_doubles
    for i in range(size):
        ri = <SparseArray> PyList_GET_ITEM(r, i)
        tmp = ytr.dot(ri)
        if not math.isfinite(tmp):
            return None        
        ri = ri * mask
        if ri.non_zero:
            PyList_Append(R, ri)
            PyList_Append(var, i)
            dependent_value[k] = tmp
            k += 1
    size = PyList_GET_SIZE(R)
    X = [array.clone(data, size+1, False) for i in range(size)]
    for i in range(size):
        data = <array.array> PyList_GET_ITEM(X, i)
        ri = <SparseArray> PyList_GET_ITEM(R, i)
        data.data.as_doubles[size] = dependent_value[i]
        for j in range(i, size):
            rj = <SparseArray> PyList_GET_ITEM(R, j)
            tmp = ri.dot(rj)
            if not math.isfinite(tmp):
                return None
            data.data.as_doubles[j] = tmp
            other = <array.array> PyList_GET_ITEM(X, j)
            other.data.as_doubles[i] = tmp
    if gauss_jordan(X):
        return None
    res = np.zeros(PyList_GET_SIZE(r))
    for k, i in enumerate(var):
        data = <array.array> PyList_GET_ITEM(X, k)
        res[i] = data[size]
    return res



# Copyright 2017 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import random
from cpython cimport array
from libc.math cimport INFINITY
from cpython.set cimport PySet_Contains, PySet_New, PySet_GET_SIZE


cdef class FunctionSelection:
    def __cinit__(self, nfunctions=0, seed=0, tournament_size=2,
                  nargs=None, density_safe=None):
        cdef unsigned int k
        self.fitness = array.clone(array.array('d'), nfunctions, zero=True)
        self.times = array.clone(array.array('I'), nfunctions, zero=True)
        self.nargs = array.clone(array.array('I'), nfunctions, zero=True)
        self.unfeasible_functions = PySet_New([])
        if nargs is not None:
            for i, k in enumerate(nargs):
                self.nargs[i] = k
        if density_safe is not None:
            self.density_safe_size = len(density_safe)
            self.density_safe = array.clone(array.array('I'), self.density_safe_size, zero=False)
            for k, d in enumerate(density_safe):
                self.density_safe[k] = d
        self.nfunctions = nfunctions
        self.tournament_size = tournament_size
        self.min_density = 0.0
        self.density = 1.0
        random.seed(seed)

    def __setitem__(self, k, v):
        self.fitness[k] += v
        self.times[k] += 1

    cpdef int random_function(self) except -1:
        cdef int value, r
        cdef Py_ssize_t _
        if self.density < self.min_density and self.density_safe_size > 0:
            for _ in range(5):
                value = random.randrange(0, self.density_safe_size)
                r = self.density_safe[value]
                if not PySet_Contains(self.unfeasible_functions, r):
                    return r
            return r
        for _ in range(5):
            r = random.randrange(0, self.nfunctions)
            if not PySet_Contains(self.unfeasible_functions, r):
                return r
        return r    

    cpdef double avg_fitness(self, Py_ssize_t x):
        cdef unsigned int _times = self.times[x]
        if _times == 0:
            return 0.0
        return self.fitness[x] / _times

    cpdef bint comparison(self, int best, int comp):
        if best == comp:
            if self.density < self.min_density:
                if self.density_safe_size == 1:
                    return False
                if PySet_GET_SIZE(self.unfeasible_functions) >= (self.density_safe_size - 1):
                    return False
            if PySet_Contains(self.unfeasible_functions, comp):
                return False
            if PySet_GET_SIZE(self.unfeasible_functions) >= (self.nfunctions - 1):
                return False
            return True
        return False

    cpdef int tournament(self) except -1:
        cdef int best, comp
        cdef double best_fit, comp_fit
        cdef unsigned int *nargs = self.nargs.data.as_uints
        if self.nfunctions == 1:
            return 0
        best = self.random_function()
        best_fit = self.avg_fitness(best)
        for i in range(1, self.tournament_size):
            comp = self.random_function()
            while self.comparison(best, comp):
                comp = self.random_function()
            comp_fit = self.avg_fitness(comp)
            if comp_fit > best_fit:
                best_fit = comp_fit
                best = comp
            elif comp_fit == best_fit and nargs[comp] > nargs[best]:
                best_fit = comp_fit
                best = comp                
        return best



# Copyright 2013 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from setuptools import setup
from Cython.Build import cythonize
from setuptools import Extension
from os.path import join
import numpy as np

with open('README.rst') as fpt:
    long_desc = fpt.read()
version = open("VERSION").readline().lstrip().rstrip()
lst = open(join("EvoDAG", "__init__.py")).readlines()
for k in range(len(lst)):
    v = lst[k]
    if v.count("__version__"):
        lst[k] = "__version__ = '%s'\n" % version
with open(join("EvoDAG", "__init__.py"), "w") as fpt:
    fpt.write("".join(lst))

extension = [Extension('EvoDAG.linalg_solve', ["EvoDAG/linalg_solve.pyx"],
                       include_dirs=[np.get_include()]),
             Extension('EvoDAG.function_selection', ["EvoDAG/function_selection.pyx"],
                       include_dirs=[np.get_include()]),
             Extension('EvoDAG.cython_utils', ["EvoDAG/cython_utils.pyx"],
                       include_dirs=[np.get_include()])]
    
setup(
    name="EvoDAG",
    description="""Evolving Directed Acyclic Graph""",
    long_description=long_desc,
    version=version,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        "Topic :: Scientific/Engineering :: Artificial Intelligence"],
    url='https://github.com/mgraffg/EvoDAG',
    author="Mario Graff",
    author_email="mgraffg@ieee.org",
    packages=['EvoDAG', 'EvoDAG/tests'],
    include_package_data=True,
    zip_safe=False,
    ext_modules=cythonize(extension,
                          compiler_directives={'profile': False,
                                               'nonecheck': False,
                                               'boundscheck': False}),

    package_data={'EvoDAG/conf': ['parameter_values.json',
                                  'default_parameters.json',
                                  'default_parameters_r.json'],
                  '': ['*.pxd']},
    install_requires=['numpy', 'SparseArray'],
    entry_points={
        'console_scripts': ['EvoDAG-params=EvoDAG.command_line:params',
                            'EvoDAG-train=EvoDAG.command_line:train',
                            'EvoDAG-predict=EvoDAG.command_line:predict',
                            'EvoDAG-utils=EvoDAG.command_line:utils'],
    }
)
